function Min(a, b)
{
  return (a < b ? a : b) + 3
}

BEGIN {

  Trim[15] = 3
  Right = "this number " left("should be equality in gender", 15)

  Right = Right " to " min(Trim[15], 2) + max(Trim[15], 3)

  print Right
}
  
